// --------------------------------------------------------------------------
//
// FishFa50.PAS                                         Ulrich Mller 1.03.02
//                                                      Procida 400NT Delphi4
//                                                      Version 5.1
// Threadfhiges FishFace auf Basis von umFish20.DLL
//    Einsatz des MultiMediaTimers um regelmige Abfrage des Interface
//    Status sicherzustellen (in umFish20.DLL)
// ft.hCom = 0 bzw. PortName = 'NONE': kein offenes Open
// und die offizielle Version : FOpen (true : erfolgreiches OpenInterface)
//
// --------------------------------------------------------------------------

unit FishFa50;

interface

uses
  Windows, MMSystem, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls;

const
  ftiAus=0; ftiEin=1; ftiLinks=1; ftiRechts=2;
  ftiOff=ftiAus; ftiOn=ftiEin; ftiLeft=ftiLinks; ftiRight=ftiRechts;
  ftiFehler=$0001FFFF; ftiError=ftiFehler;
  ftiAEX=0; ftiAEY=1; ftiEx=ftiAEX; ftiEY=ftiAEY;
  ftiE1=1; ftiE2=2; ftiE3=3; ftiE4=4; ftiE5=5; ftiE6=6; ftiE7=7; ftiE8=8;
  ftiE9=9;   ftiE10=10; ftiE11=11; ftiE12=12;
  ftiE13=13; ftiE14=14; ftiE15=15; ftiE16=16;
  ftiM1=1; ftiM2=2; ftiM3=3; ftiM4=4; ftiM5=5; ftiM6=6; ftiM7=7; ftiM8=8;
  clHellGrau = $E0E0E0;
  osWin9x = 1; osNT = 2; os2000 = 3;
  ifUndef = 0; ifLPT = 1; ifCOM = 2;
  erNone = 0; erOpen = 1; erBreak = 2;

type
  TftiDCB = record
    hCom:         LongWord;        // --- PortHandle / Openflag ( = 0)
    PortID:       LongWord;        //     IO-Adresse bei LPT
    LPTDelay:     LongWord;        //     LPT Verzgerungsfaktor
    LPTAnalog:    LongWord;        //     LPT Analogfaktor
    ECount:       LongWord;        //     Anzahl E-Eingnge (8/16)

    FID:          LongWord;        // --- MultiMediaTimer : Handle
    PollInterval: LongWord;        //     Pollinterval des MM-Timers
    AnalogScan:   LongWord;        //     Pollen auuch der Analogeingnge (0/1)

    OutputStatus: LongWord;        // --- Status der M-Ausgnge
    InputStatus:  LongWord;        //     Status der E-Eingnge
    Analogs:      array[0..1] of LongWord; //  Werte der A-Eingnge
    Counters:     array[1..16] of LongWord; // Impulse an den E-Eingngen
    end;

  TFishFace    = class;
  TPosListe     = array[1..8] of Integer;
  TFishMotList  = class;
  TFishRobot    = class;
  TfrmPanel     = class;
  TFishPosition = procedure(Sender: TFishRobot) of object;

// --------------------------TFishFace -------------------------------------

  TFishFace = class(TObject)
  private
    ft:             TftiDCB;                // --- ftiDCB von umFish.DLL
    FPortName:      string;                 // --- akt. PortName
    FOpen:          Boolean;                // --- OpenInterface erfolgreich
    FInterfaceType: Cardinal;               // --- akt. Interfacetyp
    FOSType:        Cardinal;               // --- akt. Betriebssystem
    FNotHalt:       Boolean;                // --- NotHalt
    FLastError:     LongWord;               // --- akt. festgestellter Fehler
    FAnzOutputs:    Cardinal;               // --- 4/8  mit/ohne Slave
    FVersion:       string;                 // --- Komponentenversion
    frmPanel:       TfrmPanel;              // --- InterfacePanel Form
    FPanelActive:   Boolean;
    FSleep:         LongWord;               // --- mSek Wartezeit zwischen
                                            //     den Abfragen

    function        GetAnalogWert(AnalogNr: LongWord): LongWord;
    function        GetAnalogScan: Boolean;
    function        GetSlave: Boolean;
    procedure       SetAnalogScan(Wert: Boolean);
    procedure       SetSlave(Wert: Boolean);
    function        GetCounter(InputNr: LongWord): LongWord;

  public
    property PortName: string        read FPortName;
    property LPTDelay: Cardinal      read ft.LPTDelay    write ft.LPTDelay;
    property LPTIOADR: Cardinal      read ft.PortID      write ft.PortID;
    property LPTAnalog: Cardinal     read ft.LPTAnalog   write ft.LPTAnalog;
    property Inputs: LongWord        read ft.InputStatus;
    property Counter[InputNr: LongWord]:  LongWord       read GetCounter;
    property Analogs[AnalogNr: LongWord]: LongWord       read GetAnalogWert;
    property Outputs: LongWord       read ft.OutputStatus
                                     write ft.OutputStatus;
    property NotHalt: Boolean        read FNotHalt       write FNotHalt;
    property PanelActive: Boolean    read FPanelActive;
    property LastError: LongWord     read FLastError;
    property PollInterval: LongWord  read ft.PollInterval
                                     write ft.PollInterval;
    property Slave: Boolean          read GetSlave       write SetSlave;
    property AnalogScan: Boolean     read GetAnalogScan  write SetAnalogScan;
    property ControlInterval: LongWord read FSleep write FSleep;
    property Version: string         read FVersion;

    function OpenInterface(PortName: string): LongWord;
    function CloseInterface: LongWord;

    function  GetAnalog(AnalogNr: LongWord): LongWord;
    function  GetInput(InputNr: LongWord): Boolean;
    function  GetInputs: LongWord;
    function  ClearCounter(InputNr: LongWord): LongWord;
    function  ClearCounters: LongWord;
    function  SetLamp(LampNr, OnOff: LongWord): LongWord;
    function  SetMotor(MotorNr, Direction: LongWord): LongWord;
    function  SetMotors(Status: LongWord): LongWord;
    function  ClearMotors: LongWord;
    function  Finish(InputNr: LongWord=0): Boolean;
    procedure Pause(mSek: LongWord);
    function  WaitForChange(InputNr, NrOfChanges: LongWord;
                           TermInputNr: LongWord=0): LongWord;
    function  WaitForHigh(InputNr: LongWord): LongWord;
    function  WaitForInput(InputNr: LongWord; OnOff: Boolean=true): LongWord;
    function  WaitForLow(InputNr: LongWord): LongWord;
    function  WaitForPositionDown(InputNr: LongWord; var ICounter: LongWord;
                IPosition: LongWord; TermInputNr: LongWord=0): LongWord;
    function  WaitForPositionUp(InputNr: LongWord; var ICounter: LongWord;
                IPosition: LongWord; TermInputNr: LongWord=0): LongWord;
    procedure WaitForTime(mSek: LongWord);
    function  ShowPanel(Modal: Boolean=true;
                  Links: Cardinal=300; Oben: Cardinal=300): LongWord;
    procedure ClosePanel;
    constructor Create;
    destructor  Destroy; override;
  end;

// ------------------ TFishMotList -----------------------------------------

  TMot = record
    Ende:        Cardinal;
    EndPosition: Integer;
    Impuls:      Cardinal;
    Name:        string;
    Nr:          Cardinal;
    Position:    Integer;
  end;

  TFishMotList = class(TObject)
  private
    MotorTab: array[0..7] of TMot;
    function  GetCard(i: Cardinal; Index: Integer): Cardinal;
    procedure SetCard(i: Cardinal; Index: Integer; Wert: Cardinal);
    function  GetInt(i: Cardinal;  Index: Integer): Integer;
    procedure SetInt(i: Cardinal;  Index, Wert: Integer);
    function  GetName(i: Cardinal; Index: Integer): string;
    procedure SetName(i: Cardinal; Index: Integer; Wert: string);
  public
    property Ende[i:Cardinal]: Cardinal index 1 read GetCard write SetCard;
    property EndPosition[i:Cardinal]: Integer index 2
                        read GetInt write SetInt;
    property Impuls[i:Cardinal]: Cardinal index 3
             read GetCard write SetCard;
    property Name[i:Cardinal]: string index 4 read GetName write SetName;
    property Nr[i:Cardinal]: Cardinal index 5 read GetCard write SetCard;
    property Position[i:Cardinal]: LongInt index 6
             read GetInt write SetInt;
    constructor Create;
  end;

// ----------------------------- TFishRobot --------------------------------

  TFishRobot = class(TFishFace)
  private
    FBremsweg: Integer;        // --- Anzahl Impulse
    FNachlauf: LongWord;       // --- mSek Wartezeit am Auftragsende
    FSleep: LongWord;          // --- mSek Wartezeit zwischen den Abfragen
    FOnFishPosition: TFishPosition;
    function MoveBasis(M: TPosListe; anzMot: Cardinal): LongWord;
  public
    MotCntl:  TFishMotList;
    property  BreakingDistance: Integer  read  FBremsweg write FBremsweg;
    property  StopControlTime: LongWord read  FNachlauf write FNachlauf;
    property  ControlInterval: LongWord read  FSleep    write FSleep;

    property OnFishPosition: TFishPosition read  FOnFishPosition
                                           write FOnFishPosition;

    function  MoveDelta(P1:Integer=$FFFF; P2:Integer=$FFFF; P3:Integer=$FFFF;
                        P4:Integer=$FFFF; P5:Integer=$FFFF; P6:Integer=$FFFF;
                        P7:Integer=$FFFF; P8: Integer=$FFFF): LongWord;
    function  MoveHome: LongWord;
    function  MoveTo   (P1:Integer=$FFFF; P2:Integer=$FFFF; P3:Integer=$FFFF;
                        P4:Integer=$FFFF; P5:Integer=$FFFF; P6:Integer=$FFFF;
                        P7:Integer=$FFFF; P8: Integer=$FFFF): LongWord;
    function DriveTo   (CntlNr: Cardinal; ZPos: Integer): LongWord;
    function DriveDelta(CntlNr: Cardinal; DPos: Integer): LongWord;
    function DriveFree (CntlNr: Cardinal; Direction: LongWord): LongWord;
    function DriveGrip (CntlNr: Cardinal; OnOff: LongWord): LongWord;
    constructor Create;
  end;

// --------------------------- InterfacePanel -------------------------------

  TfrmPanel = class(TForm)
    pnlButtons: TPanel;
    cmdAction: TButton;
    cmdEnde: TButton;
    lblPort: TLabel;
    shpPort: TShape;
    tmrI: TTimer;
    pnlIn1: TPanel;
    Panel1: TPanel;
    lblEingabe08: TLabel;
    lblEingabe07: TLabel;
    lblEingabe06: TLabel;
    lblEingabe05: TLabel;
    lblEingabe04: TLabel;
    lblEingabe03: TLabel;
    lblEingabe02: TLabel;
    lblEingabe01: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    pnlIn2: TPanel;
    Label11: TLabel;
    Label12: TLabel;
    Panel6: TPanel;
    lblEingabe16: TLabel;
    lblEingabe15: TLabel;
    lblEingabe14: TLabel;
    lblEingabe13: TLabel;
    lblEingabe12: TLabel;
    lblEingabe11: TLabel;
    lblEingabe10: TLabel;
    lblEingabe09: TLabel;
    pnlMot1: TPanel;
    lblMotor4: TLabel;
    Panel4: TPanel;
    lblDreh08: TLabel;
    lblDreh07: TLabel;
    lblDreh06: TLabel;
    lblDreh05: TLabel;
    lblDreh04: TLabel;
    lblDreh03: TLabel;
    lblDreh02: TLabel;
    lblDreh01: TLabel;
    lblMotor3: TLabel;
    lblMotor2: TLabel;
    lblMotor1: TLabel;
    pnlMot2: TPanel;
    lblMotor8: TLabel;
    lblMotor7: TLabel;
    lblMotor6: TLabel;
    lblMotor5: TLabel;
    Panel8: TPanel;
    lblDreh16: TLabel;
    lblDreh15: TLabel;
    lblDreh14: TLabel;
    lblDreh13: TLabel;
    lblDreh12: TLabel;
    lblDreh11: TLabel;
    lblDreh10: TLabel;
    lblDreh09: TLabel;
    lblEHex: TLabel;
    lblAHex: TLabel;
    Label47: TLabel;
    Label48: TLabel;
    pnlAnalog: TPanel;
    Label53: TLabel;
    Label52: TLabel;
    lblAnalogY: TLabel;
    Label51: TLabel;
    lblAnalogX: TLabel;
    procedure cmdEndeClick(Sender: TObject);
    procedure cmdActionClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure tmrITimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure lblDrehDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lblDrehUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lblMotorClick(Sender: TObject);
  private
    MotorDauer: array[1..8] of Boolean;
    AusgabeLabel: array[1..16] of TLabel;
    EingabeLabel: array[1..16] of TLabel;
    Vater: TFishFace;
    procedure MotorAusS(Mnr: Cardinal);
  public
    { Public-Deklarationen }
  end;

// --------------------------------- implementation -------------------------

implementation

{$R *.DFM}

const
  vbTrue=-1; vbFalse=0;

  MAus:    array[1..8]  of LongWord  =     ($FFFC, $FFF3, $FFCF, $FF3F,
                                            $FCFF, $F3FF, $CFFF, $3FFF);
  MLinks:  array[1..8]  of LongWord  =     ($1,    $4,    $10,   $40,
                                            $100,  $400,  $1000, $4000);
  MRechts: array[1..8]  of LongWord  =     ($2,    $8,    $20,   $80,
                                            $200,  $800,  $2000, $8000);
  EMaske:  array[1..16] of LongWord =      ($1,    $2,    $4,    $8,
                                            $10,   $20,   $40,   $80,
                                            $100,  $200,  $400,  $800,
                                            $1000, $2000, $4000, $8000);

function umOpenInterface(var ft: TftiDCB; PortName: string):LongInt;
       stdcall; external 'umFish20.dll';
function umCloseInterface(var ft: TftiDCB):LongInt;
       stdcall; external 'umFish20.dll';
function umGetVersion:LongInt;
       stdcall; external 'umFish20.dll';
function umGetInput(var ft: TftiDCB; InputNr:LongInt):LongInt;
       stdcall; external 'umFish20.dll';
function umSetMotor(var ft: TftiDCB; MotorNr, Direction:LongInt):LongInt;
       stdcall; external 'umFish20.dll';
function umSetLamp(var ft: TftiDCB; LampNr, OnOff:LongInt):LongInt;
       stdcall; external 'umFish20.dll';
function umStartDriver:LongInt; stdcall; external 'umFish20.dll';
function umStopDriver:LongInt;  stdcall; external 'umFish20.dll';



// ---- div Routinen ------------------------------------------------------

function ChkPortName(N: string): Cardinal;
// --- Prfen auf korrekten PortNamen, Rckgabe FInterfaceType ------------
begin
  Result := ifUndef;
  N := UpperCase(N);
  if (length(N) = 3) AND (N = 'LPT') then begin Result := ifLPT; exit end;
  if length(N) <> 4 then exit;
  if NOT (N[4] in ['1'..'9']) then exit;
  if (Copy(N,1,3) = 'COM') then begin Result := ifCOM; exit end;
  if (Copy(N,1,3) <> 'LPT') then exit;
  if NOT (N[4] in ['1'..'3']) then exit;
  Result := ifLPT;
end;



// -------------------- { TFishFace } --------------------------------------

function TFishFace.ClearCounter(InputNr: LongWord): LongWord;
// --- Lschen ImpulsCounter ---------------------------------
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  ft.Counters[InputNr] := 0;
  Result := 0;
end;


function TFishFace.ClearCounters: LongWord;
// --------------- Lschen aller ImpulsCounter ---------------

var i: Cardinal;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i := 1 to ft.ECount do ft.Counters[i] := 0;
  Result := 0;
end;


function TFishFace.ClearMotors: LongWord;
// ------- Lschen aller M-Ausgnge -------------------------
begin
  ft.OutputStatus := 0;
  Result   := ft.InputStatus;
end;


function TFishFace.CloseInterface: LongWord;
// ----- Schlieen Interface ---------------------------------
begin
  Result := ftiFehler;
  if ft.hCom <> 0 then begin
    ft.OutputStatus := 0;
    Sleep(4*ft.PollInterval);          // --- Zeit fr PollInterface um
                                       //     FOutputs auszufhren
    if ft.FID <> 0 then timekillEvent(ft.FID);
    Result := umCloseInterface(ft);
    FOpen := false;
    if Result = ftiFehler then FLastError := erOpen;
    FPortName := 'NONE';
    FOpen := false;
    end;
end;


constructor TFishFace.Create;
var OS: TOSVersionInfo;
begin
  inherited Create;
  FPortName       := 'NONE';
  FOpen           := false;
  FInterfaceType  := 0;
  ft.ECount       := 8;
  FAnzOutputs     := 4;
  ft.AnalogScan   := 0;
  ft.InputStatus  := 0;
  ft.Analogs[0]   := 0;
  ft.Analogs[1]   := 0;
  ft.OutputStatus := 0;
  FNotHalt        := false;
  FPanelActive    := false;
  FLastError      := erOpen;
  ft.PollInterval := 10;
  FSleep          := 10;
  FVersion        := '5.1';
  FOSType         := osWin9x;
  OS.dwOSVersionInfoSize := sizeof(OS);
  GetVersionEX(OS);
  if OS.dwPlatformId = Ver_Platform_Win32_NT then FOSType := osNT;
end;


destructor TFishFace.Destroy;
begin
  if ft.hCom <> 0 then CloseInterface;
  inherited Destroy;
end;


function TFishFace.Finish(InputNr: LongWord=0): Boolean;
// --- Feststellen eines EndeWunsches : NotHalt, Escape, InputNr ---
begin
  Result := true;
  Application.ProcessMessages;
  if NOT FOpen then exit;
  if FNotHalt OR (GetAsyncKeyState(VK_ESCAPE) <> 0) then exit;
  if InputNr > 0 then if GetInput(InputNr) then exit;
  Result := false;
end;


function TFishFace.GetAnalog(AnalogNr: LongWord): LongWord;
// --- Auslesen Analogwert, Methode ---------------------------------
begin
  Result := ftiFehler;
  Application.ProcessMessages;
  if NOT FOpen then exit;
  Result := ft.Analogs[AnalogNr];
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


function TFishFace.GetAnalogScan: Boolean;
// ------- Pollen mit/ohne Analogwerte ----
begin
  if ft.AnalogScan  = 0 then Result := false else Result := true;
end;


function TFishFace.GetAnalogWert(AnalogNr: LongWord): LongWord;
// --- letzter A-Wert ------------------------------------------
begin
  Result := ft.Analogs[AnalogNr];
end;


function TFishFace.GetCounter(InputNr: LongWord): LongWord;
// --- Auslesen ImpulsCounter ------------------------------
begin
  Result := ft.Counters[InputNr];
end;


function TFishFace.GetInput(InputNr: LongWord): Boolean;
// -------------- Bestimmen Status E-Eingng --------------
begin
  Result := false;
  Application.ProcessMessages;
  if NOT FOpen then exit;
  if ft.InputStatus = ftiFehler then FLastError := ftiFehler
  else FLastError := erNone;
  if (ft.InputStatus AND EMaske[InputNr]) > 0 then Result := true;
end;


function TFishFace.GetInputs: LongWord;
// --- Auslesen Status E-Eingnge ----------------------------
begin
  Result := ftiFehler;
  Application.ProcessMessages;
  if NOT FOpen then exit;
  Result := ft.InputStatus;
  if Result = ftiFehler then FLastError := ftiFehler
  else FLastError := erNone;
end;


function TFishFace.GetSlave: Boolean;
// --- Eigenschaft : mit/ohne Slave ----------------------------
begin
  if FAnzOutputs = 4 then Result := false else Result := true;
end;


function TFishFace.OpenInterface(PortName: string): LongWord;
// ----- OpenInterface, Bestimmen von FPollInterval ------------
//       Starten MultiMediaTimer -------------------------------
begin
  Result := ftiFehler; FPortName := 'NONE';
  FOpen := false; FLastError := erOpen;

  FInterfaceType := ChkPortName(PortName);
  if FInterfaceType = ifUndef then exit;
  Result := umOpenInterface(ft, PortName);
  if Result = ftiFehler then exit;
  FPortName := PortName; ft.OutputStatus := 0; FNotHalt := false;
  FOpen := true; FLastError := erNone;
end;


procedure TFishFace.Pause(mSek: LongWord);
// --- Eine vorgegebene Zeit warten ----------------
//     Abbruch beit NotHalt, Escape ----------------

var sZeit: DWORD;

begin
  sZeit := GetTickCount;
  repeat
    Application.ProcessMessages;
    Sleep(ft.PollInterval);
    if ft.InputStatus = ftiFehler then begin
      FLastError := ftiFehler;
      exit;
      end;
    if FNotHalt OR (GetAsyncKeyState(VK_ESCAPE) <> 0) then begin
      FLastError := erBreak;
      exit;
      end;
  until (GetTickCount -sZeit > mSek) ;
end;


procedure TFishFace.SetAnalogScan(Wert: Boolean);
// --- Eigenschaft : mit/ohne Auslesen A-Eingnge -----
begin
  if Wert then ft.AnalogScan := 1 else ft.AnalogScan := 0;
end;


function TFishFace.SetLamp(LampNr, OnOff: LongWord): LongWord;
// ------ setzen eines 'halben' M-Ausganges -------------------
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  Result   := umSetLamp(ft, LampNr, OnOff);
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


function TFishFace.SetMotor(MotorNr, Direction: LongWord): LongWord;
// ---- Setzen Status M-Ausgang -------------------------------------
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  Result := umSetMotor(ft, MotorNr, Direction);
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


function TFishFace.SetMotors(Status: LongWord): LongWord;
// --- Setzen aller M-Ausgnge ---------------------------
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  ft.OutputStatus := Status;
  Result   := ft.InputStatus;
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


procedure TFishFace.SetSlave(Wert: Boolean);
// --- Eigenschaft mit/ohne Slave --------------
begin
  if Wert then begin
    ft.ECount   := 16;
    FAnzOutputs := 8;
    end
  else begin
    ft.ECount   := 8;
    FAnzOutputs := 4;
    end;
end;

function TFishFace.ShowPanel(Modal: Boolean=true;
           Links: Cardinal=300; Oben: Cardinal=300):LongWord;
// --- Anzeige InterfacePanel -------------------------------------------------
begin
  Result         := ftiFehler;
  if NOT FOpen then exit;
  frmPanel       := TfrmPanel.Create(Application.MainForm);
  frmPanel.Vater := Self;
  frmPanel.Left  := Links;
  frmPanel.Top   := Oben;
  FPanelActive   := true;
  if Modal then begin
    frmPanel.ShowModal;
    FPanelActive := false;
    end
  else frmPanel.Show;
  Result         := 0;
end;

procedure TFishFace.ClosePanel;
// --- Schlieen InterfacePanel ----------------------------------------
begin
  if NOT FPanelActive then exit;
  frmPanel.Close;
  FPanelActive := false;
end;

function TFishFace.WaitForChange(InputNr, NrOfChanges, TermInputNr: LongWord): LongWord;
// ---- Warten auf NrOfChanges Impuls von InputNr ---------------------------
//      Abbruch : NotHalt, Escape, ftiFehler oder TermInputNr

var M3: LongWord;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  ft.Counters[InputNr] := 0;

  if TermInputNr = 0 then M3 := 0
  else begin
    M3 := EMaske[TermInputNr];
    if ft.InputStatus = ftiFehler then begin
      FLastError := ftiFehler; exit; end;
    if (ft.InputStatus AND M3) > 0 then begin
      FLastError := erNone; Result := ft.InputStatus; exit; end;
    end;

  while ft.Counters[InputNr] < NrOfChanges do begin
    if FNotHalt OR (GetAsyncKeyState(VK_ESCAPE) <> 0) then begin
      FLastError := erBreak; exit; end;
    if ft.InputStatus = ftiFehler then begin
      FLastError := ftiFehler; exit; end;
    if (ft.InputStatus AND M3) > 0 then break;
    Application.ProcessMessages;
    Sleep(FSleep);
    end;
  Result := ft.InputStatus;
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;

function  TFishFace.WaitForHigh(InputNr: LongWord): LongWord;
// --- Warten auf einen false/true Durchgang von InputNr ---
//     Abbruch bei NotHalt, Escape, Fehler
begin
  WaitForInput(InputNr, false);
  Result := WaitForInput(InputNr, true);
end;

function TFishFace.WaitForInput(InputNr: LongWord; OnOff: Boolean=true): LongWord;
// --- Warten auf InputNr = OnOff ------------------------------------------
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  repeat
    if FNotHalt OR (GetAsyncKeyState(VK_ESCAPE) <> 0) then begin
      FLastError := erBreak; exit; end;
    if ft.InputStatus = ftiFehler then begin
      FLastError := ftiFehler; exit; end;
    Application.ProcessMessages;
    Sleep(FSleep);
  until ((ft.InputStatus AND EMaske[InputNr]) > 0) = OnOff;
  Result := ft.InputStatus;
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


function TFishFace.WaitForLow(InputNr: LongWord): LongWord;
// --- Warten auf Wechsel von InputNr auf ftiAus ----------------------------
begin
  WaitForInput(InputNr, true);
  Result := WaitForInput(InputNr, false);
end;

function  TFishFace.WaitForPositionDown(InputNr: LongWord; var ICounter: LongWord;
            IPosition: LongWord; TermInputNr: LongWord=0): LongWord;
// --- Warten auf IPosition = ICounter durch Subtrahieren der Impulse
//     des Einganges InputNr bzw, TermInputNr = true
//     Abbruch bei NotHalt, Escape, Fehler, Endtaster
begin
  Result := WaitForChange(InputNr, ICounter - IPosition, TermInputNr);
  ICounter := ICounter - ft.Counters[InputNr];
end;

function  TFishFace.WaitForPositionUp(InputNr: LongWord; var ICounter: LongWord;
            IPosition: LongWord; TermInputNr: LongWord=0): LongWord;
// --- Warten auf IPosition = ICounter durch Addieren der Impulse
//     des Einganges InputNr bzw. TermInputNr = true
//     Abbruch bei NotHalt, Escape, Fehler, Endtaster
begin
  Result := WaitForChange(InputNr, IPosition - ICounter, TermInputNr);
  ICounter := ICounter + ft.Counters[InputNr];
end;

procedure TFishFace.WaitForTime(mSek: LongWord);
// --- Eine vorgegebene Zeit Warten -----
//     Abbruch bei NotHalt, Escape
//     Synonym fr Pause aus Kompatibilittsgrnden
begin
  Pause(mSek);
end;


// TFishMotorList-Methoden ----------------------------------------------------
//
// Liste von simultan zu steuernden Motoren
//
// -------------------------------------------------------------------------

constructor TFishMotList.Create;
var i: Cardinal;
begin
  inherited Create;
  for i := 0 to 7 do begin
    MotorTab[i].Ende        := i * 2 + 1;
    MotorTab[i].EndPosition := 9999;
    MotorTab[i].Impuls      := i * 2 + 2;
    MotorTab[i].Name        := 'Dummy';
    MotorTab[i].Nr          := i + 1;
    MotorTab[i].Position    := 9999;
    end;
end;

function TFishMotList.GetCard(i: Cardinal; Index: Integer): Cardinal;
begin
  with MotorTab[i-1] do
    case Index of
    1: Result := Ende;
    3: Result := Impuls;
    5: Result := Nr;
    else
      Result := $FFFF;
    end;
end;

procedure TFishMotList.SetCard(i: Cardinal; Index: Integer; Wert: Cardinal);
begin
  with MotorTab[i-1] do
    case Index of
    1: Ende        := Wert;
    3: Impuls      := Wert;
    5: Nr          := Wert;
    end;
  end;

function TFishMotList.GetInt(i: Cardinal; Index: Integer): Integer;
begin
  with MotorTab[i-1] do
    case Index of
    2: Result := EndPosition;
    6: Result := Position;
    else
      Result := $FFFF;
    end;
end;

procedure TFishMotList.SetInt(i: Cardinal; Index, Wert: Integer);
begin
  with MotorTab[i-1] do
    case Index of
    2: EndPosition := Wert;
    6: Position    := Wert;
    end;
  end;

function TFishMotList.GetName(i: Cardinal; Index: Integer): string;
begin
  GetName := MotorTab[i-1].Name;
end;

procedure TFishMotList.SetName(i: Cardinal; Index: Integer; Wert: string);
begin
  MotorTab[i-1].Name := Wert;
end;



// TFishRobot --------------------------------------------------------------
//
// Funktionen zum simultanen Fahren mehrerer Motoren
//                    zu einer vorgegebenen Position
// Anfahren Endtaster : "linksdrehend" (ftiLinks)
//
// -------------------------------------------------------------------------

constructor TFishRobot.Create;
begin
  inherited Create;
  MotCntl          := TFishMotList.Create;
  FBremsweg := 6;
  FNachlauf := 100;   // --- Dauer von 2,5 Impulsen ---------
  FSleep    := 10;
end;

function TFishRobot.DriveDelta(CntlNr: Cardinal; DPos: Integer): LongWord;
// --- Fahren eines einzelnen Motor aus MotCntl, Differenz Angabe --------
var M: TPosListe; i: LongWord;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i:= 1 to CntlNr do
    if i = CntlNr then M[i] := MotCntl.Position[i] + DPos
    else M[i] := MotCntl.Position[i];
  Result := MoveBasis(M, CntlNr);
end;

function TFishRobot.DriveFree(CntlNr: Cardinal;
  Direction: LongWord): LongWord;
// --- Freies Fahren eines Motors, Ende durch NotHalt ---------------------
//     die Grenzen 0 - EndPosition werden beachtet
var M: TPosListe; i: LongWord;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i := 1 to CntlNr do
    if i = CntlNr then if Direction = ftiLinks then M[i] := 0
                       else M[i] := MotCntl.EndPosition[i]
    else M[i] := MotCntl.Position[i];
  Result := MoveBasis(M, CntlNr);
end;

function TFishRobot.DriveGrip(CntlNr: Cardinal; OnOff: LongWord): LongWord;
// --- ffnen/Schlieen eines Greifer aus MotCntl -------------------------
var M: TPosListe; i: LongWord;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i := 1 to CntlNr do
    if i = CntlNr then if OnOff = ftiAus then M[i] := 0
                       else M[i] := MotCntl.EndPosition[i]
    else M[i] := MotCntl.Position[i];
  Result := MoveBasis(M, CntlNr);
end;

function TFishRobot.DriveTo(CntlNr: Cardinal; ZPos: Integer): LongWord;
// --- Fahren eines einzelnen Motor aus MotCntl ---------------------------
var M: TPosListe; i: LongWord;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i:= 1 to CntlNr do
    if i = CntlNr then M[i] := ZPos else M[i] := MotCntl.Position[i];
  Result := MoveBasis(M, CntlNr);
end;

function TFishRobot.MoveBasis(M: TPosListe; anzMot: Cardinal): LongWord;
// --- Simultanes Fahren der Motoren zu den in M angegebenen Positionen ----
var
  MotDreh, MotAus, MotEnde, i: LongWord;
  MaskeEingaenge: LongWord;
  N: array[1..8] of Integer; EventZaehler: LongWord;
  BremsPhase: array[1..8] of Integer;
  Bremse: array[1..8] of Boolean;
begin
  Result := ftiFehler;
  if NOT FOPen then exit;
  EventZaehler := 0;
  MaskeEingaenge := 0;
  MotDreh        := ft.OutputStatus;
  MotAus         := MotDreh;

with MotCntl do begin
// --- Fahrauftrag und Masken bestimmen -----------------------------------

  for i := 1 to anzMot do begin
    MotDreh   := MotDreh AND MAus[Nr[i]];  // --- EinzelMotor aus
    MotAus    := MotAus  AND MAus[Nr[i]];  // --- Alle Motoren aus
    ft.Counters[Impuls[i]] := 0;                // --- Counter zurck
    Bremse[i] := false;

    MaskeEingaenge := MaskeEingaenge OR EMaske[Impuls[i]] OR EMaske[Ende[i]];

// ---- Fahrauftrag fr Einzelmotor -----------------------------------------

    if M[i] > Position[i] then begin       // --- rechts : weg vom Endtaster
      if M[i] > EndPosition[i] then M[i] := EndPosition[i];
      MotDreh := MotDreh OR MRechts[Nr[i]];
      BremsPhase[i] := M[i] - FBremsweg;
      end;
                                           // --- links : zum Endtaster
    if (ft.InputStatus AND EMaske[Ende[i]]) > 0 then Position[i] := 0
    else begin
      if (M[i] <= 0) AND (Position[i] > 0) then M[i] := -9999;
      if M[i] < Position[i] then begin
        MotDreh := MotDreh OR MLinks[Nr[i]];
        BremsPhase[i] := M[i] + FBremsweg;
        end;
      end;
    N[i] := Position[i];
    end;
  MotEnde := MotDreh;

// --- Motoren starten -----------------------------------------

  ft.OutputStatus := MotDreh;

//
// ---- Schleife ber Fahrauftrag ---------------------------------------
//

  while MotEnde <> MotAus do begin
    if FNotHalt OR (GetAsyncKeyState(VK_ESCAPE) <> 0) then begin
      ft.OutputStatus := MotAus;
      Result := ftiFehler;
      FLastError := erBreak;
      exit;
      end
    else if ft.InputStatus = ftiFehler then begin
      ft.OutputStatus := MotAus;
      Result := ftiFehler;
      FLastError := ftiFehler;
      exit;
      end;

//  --- Fahrauftrag und Status aktualisieren -----------------------------

    Sleep(FSleep);                  // --- Abfrage alle ControlInterval --
    Application.ProcessMessages;
    ft.OutputStatus := MotDreh;

//  --- Bearbeiten der Vernderungen -------------------------------------

    for i := 1 to anzMot do begin

      if (MotEnde AND MRechts[Nr[i]]) > 0 then begin
        // --- rechts : weg vom Taster -------------
        Position[i] := N[i] +  Integer(ft.Counters[Impuls[i]]);
        if (Position[i] >= BremsPhase[i]) then
          if Bremse[i] then begin                   // --- Bremsen ------
            MotDreh := MotDreh OR MRechts[Nr[i]]; Bremse[i] := false; end
          else begin
            MotDreh := MotDreh AND MAus[Nr[i]]; Bremse[i] := true; end;
        if Position[i] >= M[i] then begin
          // --- Motor ausschalten ----------------------------------
          MotDreh := MotDreh AND MAus[Nr[i]]; MotEnde := MotDreh; end;
        end

      else if (MotEnde AND MLinks[Nr[i]]) > 0 then begin
        // --- links : hin zum Taster -------------------
        Position[i] := N[i]- Integer(ft.Counters[Impuls[i]]);
        if (Position[i] <= BremsPhase[i]) then
          if Bremse[i] then begin                      // --- Bremsen ---
            MotDreh := MotDreh OR MLinks[Nr[i]]; Bremse[i] := false; end
          else begin
            MotDreh := MotDreh AND MAus[Nr[i]]; Bremse[i] := true; end;
        if Position[i] <= M[i] then begin
          // --- Motor ausschalten -----------------------------------
          MotDreh := MotDreh AND MAus[Nr[i]]; MotEnde := MotDreh; end;
        if (ft.InputStatus AND EMaske[MotCntl.Ende[i]]) > 0 then begin
          // ---- Normieren auf Position 0 --------------------
          Position[i] := 0;
          MotDreh := MotDreh AND MAus[Nr[i]];
          MotEnde := MotDreh;
          end;
        end;
      end;  // --- for anzMot
      if Assigned(FOnFishPosition) then FOnFishPosition(Self);
    end;  // --- while MotEnde

// ---- Update Position ----------------------------------------------

  ft.OutputStatus := MotAus;
  Sleep(FNachlauf); // --- Abwarten Nachlauf
  for i := 1 to anzMot do
    if M[i] > Position[i] then
      Position[i] := N[i] + Integer(ft.Counters[Impuls[i]])
    else if M[i] < Position[i] then
      Position[i] := N[i] - Integer(ft.Counters[Impuls[i]]);
  if EventZaehler = 5 then EventZaehler := 0 else Inc(EventZaehler);
  if EventZaehler = 5 then if Assigned(FOnFishPosition) then
    FOnFishPosition(Self);
  Application.ProcessMessages;

end;  // --- with MotCntl

  Result := ft.InputStatus;
  if Result = ftiFehler then FLastError := ftiFehler else FLastError := erNone;
end;


function TFishRobot.MoveDelta(P1, P2, P3, P4, P5, P6, P7,
                              P8: Integer): LongWord;
// --- Fahren auf Position rel. zum Ursprung (Endtaster) -------------------
var M: TPosListe; anzMot: Cardinal;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  anzMot := 8;
with MotCntl do begin
  if P8 = $FFFF then Dec(anzMot) else M[8] := Position[8] + P8;
  if P7 = $FFFF then Dec(anzMot) else M[7] := Position[7] + P7;
  if P6 = $FFFF then Dec(anzMot) else M[6] := Position[6] + P6;
  if P5 = $FFFF then Dec(anzMot) else M[5] := Position[5] + P5;
  if P4 = $FFFF then Dec(anzMot) else M[4] := Position[4] + P4;
  if P3 = $FFFF then Dec(anzMot) else M[3] := Position[3] + P3;
  if P2 = $FFFF then Dec(anzMot) else M[2] := Position[2] + P2;
  if P1 = $FFFF then exit        else M[1] := Position[1] + P1;
end;
  Result := MoveBasis(M, anzMot);
end;



function TFishRobot.MoveHome: LongWord;
// --------------- Simultanes Fahren auf die HomePosition -------------------
var i: Cardinal; EPos: array[1..8] of Integer;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  for i := 1 to 8 do begin
    EPos[i] := $FFFF;        // --- Motor nicht vorhanden
    if MotCntl.EndPosition[i] < 9999 then EPos[i] := -9999;
  end;
  Result := MoveTo(EPos[1],EPos[2], EPos[3], EPos[4],
                   EPos[5], EPos[6], EPos[7], EPos[8]);
  for i := 1 to 8 do MotCntl.Position[i] := 0;
end;



function  TFishRobot.MoveTo(P1: Integer=$FFFF; P2: Integer=$FFFF;
                         P3: Integer=$FFFF; P4: Integer=$FFFF;
                         P5: Integer=$FFFF; P6: Integer=$FFFF;
                         P7: Integer=$FFFF; P8: Integer=$FFFF): LongWord;
// --------------- Simultanes Fahren auf die angegebene Postision ----------
var M: TPosListe; anzMot: Cardinal;
begin
  Result := ftiFehler;
  if NOT FOpen then exit;
  anzMot := 8;
  if P8 = $FFFF then Dec(anzMot) else M[8] := P8;
  if P7 = $FFFF then Dec(anzMot) else M[7] := P7;
  if P6 = $FFFF then Dec(anzMot) else M[6] := P6;
  if P5 = $FFFF then Dec(anzMot) else M[5] := P5;
  if P4 = $FFFF then Dec(anzMot) else M[4] := P4;
  if P3 = $FFFF then Dec(anzMot) else M[3] := P3;
  if P2 = $FFFF then Dec(anzMot) else M[2] := P2;
  if P1 = $FFFF then exit        else M[1] := P1;
  Result := MoveBasis(M, anzMot);
end;



// --- frmPanel --------------------------------------------------------------
//
// InterfacePanel. Erstmals mit variabler Gre in Abhngigkeit vom Slave
//
// ---------------------------------------------------------------------------

procedure TfrmPanel.cmdEndeClick(Sender: TObject);
// --- Schlieen InterfacePanel ---------------------------------------------
begin
  Vater.FPanelActive := false;
  Application.ProcessMessages;
  Self.Close;
end;

procedure TfrmPanel.cmdActionClick(Sender: TObject);
// --------------- Lschen aller Ausgnge ----------------------------------
begin
  Vater.NotHalt := true;
  Vater.ClearMotors;
end;

procedure TfrmPanel.FormCreate(Sender: TObject);
// ----------- Besetzen der Labelzuordnungtabellen ------------------------
begin
  AusgabeLabel[01] := lblDreh01;
  AusgabeLabel[02] := lblDreh02;
  AusgabeLabel[03] := lblDreh03;
  AusgabeLabel[04] := lblDreh04;
  AusgabeLabel[05] := lblDreh05;
  AusgabeLabel[06] := lblDreh06;
  AusgabeLabel[07] := lblDreh07;
  AusgabeLabel[08] := lblDreh08;
  AusgabeLabel[09] := lblDreh09;
  AusgabeLabel[10] := lblDreh10;
  AusgabeLabel[11] := lblDreh11;
  AusgabeLabel[12] := lblDreh12;
  AusgabeLabel[13] := lblDreh13;
  AusgabeLabel[14] := lblDreh14;
  AusgabeLabel[15] := lblDreh15;
  AusgabeLabel[16] := lblDreh16;

  EingabeLabel[01] := lblEingabe01;
  EingabeLabel[02] := lblEingabe02;
  EingabeLabel[03] := lblEingabe03;
  EingabeLabel[04] := lblEingabe04;
  EingabeLabel[05] := lblEingabe05;
  EingabeLabel[06] := lblEingabe06;
  EingabeLabel[07] := lblEingabe07;
  EingabeLabel[08] := lblEingabe08;
  EingabeLabel[09] := lblEingabe09;
  EingabeLabel[10] := lblEingabe10;
  EingabeLabel[11] := lblEingabe11;
  EingabeLabel[12] := lblEingabe12;
  EingabeLabel[13] := lblEingabe13;
  EingabeLabel[14] := lblEingabe14;
  EingabeLabel[15] := lblEingabe15;
  EingabeLabel[16] := lblEingabe16;
end;

procedure TfrmPanel.FormShow(Sender: TObject);
// --- Rearrange bei kleinem Interface, Starten Timer ------------------
begin
  if NOT Vater.Slave then begin
    pnlIn2.Visible  := false;
    pnlMot2.Visible := false;
    pnlAnalog.Left := lblAHex.Left;
    pnlButtons.Left := lblAHex.Left;
    pnlIn1.Left  := pnlAnalog.Left + pnlAnalog.Width - pnlIn1.Width;
    pnlMot1.Left := pnlAnalog.Left + pnlAnalog.Width - pnlMot1.Width;
    Width :=  330;
  end;
  lblPort.Caption := Vater.PortName;
  tmrI.Enabled := true;
end;

procedure TfrmPanel.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  tmrI.Enabled := false;         // --- sonst kein Ende
end;

procedure TfrmPanel.tmrITimer(Sender: TObject);
// --- Zyklische Anzeige des Interface-Status ----------------------------
var
  i, OG: Cardinal; w: string;
begin
  if Vater.Inputs = ftiFehler then begin
    shpPort.Brush.Color := clRed;
    exit;
    end
  else shpPort.Brush.Color := clLime;

  Str(Vater.GetAnalog(0):4, w);
  lblAnalogX.Caption := w;
  Str(Vater.GetAnalog(1):4, w);
  lblAnalogY.Caption := w;

  if Vater.Slave then OG := 16 else OG := 8;
  for i := 1 to OG do begin
    if (Vater.Inputs AND EMaske[i]) <> 0 then
      EingabeLabel[i].Color := clLime
    else EingabeLabel[i].Color := clBtnFace;
    if (Vater.Outputs AND EMaske[i]) <> 0 then
      AusgabeLabel[i].Color := clLime
    else AusgabeLabel[i].Color := clBtnFace;
    if Vater.ft.ECount = 16 then begin
      lblAHex.Caption := IntToHex(Vater.Outputs,4);
      lblEHex.Caption := IntToHex(Vater.Inputs,4);
      end
    else begin
      lblAHex.Caption := IntToHex((Vater.Outputs AND $FF),4);
      lblEHex.Caption := IntToHex((Vater.Inputs AND $FF),4);
      end;
    end;
end;

procedure TfrmPanel.MotorAusS(Mnr: Cardinal);
// ---- Ausschalten eines MotorAusgangen ----------------------------------
begin
  Vater.SetMotor(Mnr, ftiAus);
  AusgabeLabel[(Mnr-1)*2+1].Color := clBtnFace;
  AusgabeLabel[(Mnr-1)*2+2].Color := clBtnFace;
  lblAHex.Caption := IntToHex(Vater.Outputs,4);
end;

procedure TfrmPanel.lblDrehDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
// --- Motorausgang schalten ------------------------------------------------
//     solange MousDown, wenn Strg gedrck : Dauer
var
  Index, Ri, Mnr: Cardinal;
begin
  Index := TLabel(Sender).Tag;
  if (Index mod 2) = 0 then Ri := ftiRechts else Ri := ftiLinks;
  Mnr := (Index-1) div 2 + 1;
  if ssCtrl in Shift then MotorDauer[Mnr] := true;
  Vater.SetMotor(Mnr, Ri);
  TLabel(Sender).Color := clLime;
  lblAHex.Caption := IntToHex(Vater.Outputs,4);
end;

procedure TfrmPanel.lblDrehUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
// --- Motorausgang abschalten, wenn kein Dauer ---------------------------
var
  Index, Mnr: Cardinal;
begin
  Index := TLabel(Sender).Tag;
  Mnr := (Index-1) div 2 + 1;
  if MotorDauer[Mnr] then exit;
  MotorAusS(Mnr);
end;

procedure TfrmPanel.lblMotorClick(Sender: TObject);
// --- Motorausgang abschalten, wenn Dauer --------------------
var
  Index: Cardinal;
begin
  Index := TLabel(Sender).Tag;
  MOtorDauer[Index] := false;
  MotorAusS(Index);
end;

end.
